/*  showcomm.c : ver 1.2 : 24/01/96  */

/* 
 *  (C.) Copyright 1995 University of Liverpool
 *
 *  Author: d.j.nixon@csc.liv.ac.uk
 */

/*
 *   Usage: showcomm <username>|all  <upacct file>
 *          whowcomm <username>      <upacct file>
 *
 *   "showcomm" Lists when, where and for how long selected users ran commands
 *    - the "all" option lists the totals for all user commands.
 *   "whocomm" Lists all users who ran a specified command.
 */  

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <pwd.h>
#include <errno.h>
#include <time.h>
#include <sys/acct.h>

#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>

/* global data */
extern char *sys_errlist[];

/* Functions */
extern char* get_cmd_from_path();

#define COMMLEN 8        /* max length of command names */

/* list of per user data */
typedef struct uacct {
  time_t ac_btime;       /* Beginning time */
  comp_t ac_etime;       /* acctng elapsed time in clock ticks */
  char ac_comm[COMMLEN];       /* command name */
  struct in_addr addr;   /* host address */
  struct uacct *next;
} uacct;

/* upacct file record header */
typedef struct uahdr {
u_short uid;
unsigned count;
} uahdr;

main(argc, argv)
int argc;
char **argv;
{
uacct uacctbuf;
uahdr hdr;
int upacct;
register int bytes,  tflag=0, cflag=0, len;
register u_short uid;
struct passwd *pwd;
struct hostent *hp;
char user_id[16], *uname, *cmd=get_cmd_from_path(argv[0]);

  /* parse command line */
  if (!strcmp(cmd, "whocomm"))
  {
    if (argc < 3)
    {
      fputs("Usage: whocomm <command>  <upacct file>\n",stderr);
      exit(1);
    }
    len=strlen(argv[1]);
    if (len > COMMLEN)
    {
      len=COMMLEN;
      fprintf(stderr,"Warning: input data doesn't distinguish command names longer than %u characters\n",COMMLEN);
    }
    ++cflag;
  }
  else if (!strcmp(cmd, "showcomm"))
  {
    if (argc < 3)
    {
      fputs("Usage: showcomm <username>|all  <upacct file>\n",stderr);
      exit(1);
    }
    if (!strcmp(argv[1], "all"))
      ++tflag;                                  /* all users totals option */
    else if ((pwd=getpwnam(argv[1])) == NULL)
    {
      fprintf(stderr,"No passwd file record for %s !\n",argv[1]);
      exit(1);
    }
    else
      uid=pwd->pw_uid;                         /* default is user search */
  }
  else
  {
    fprintf(stderr,"Cannot be run as %s\n",cmd);
    exit(1);
  }

  /* open command accounting data file */
  if ((upacct=open(argv[2], O_RDONLY)) == -1)
  {
    fprintf(stderr,"%s: can't open %s: %s\n",cmd,argv[2],sys_errlist[errno]);
    exit(1);
  }

  /*
   *  Display user/command information.
   */
  uacctbuf.next=NULL;
  while(bytes=read(upacct, &hdr, sizeof(uahdr)))
  {
    if (bytes == -1)
    {
      fprintf(stderr,"%s: error reading %s: %s\n",cmd,argv[1],sys_errlist[errno]);
      exit(1);
    }
    if (tflag)
    {
      if ((pwd=getpwuid(hdr.uid)) == NULL)
        printf("%u - %u\n",hdr.uid,hdr.count);
      else
        printf("%s - %u\n",pwd->pw_name,hdr.count);
    }
    else if ((hdr.uid ==  uid)  || (cflag))
    {
      if (cflag)  
      {
        if ((pwd=getpwuid(hdr.uid)) == NULL)
        {
          sprintf(user_id,"%u",hdr.uid);
          uname=user_id;
        }
        else
          uname=pwd->pw_name;
      }
      while(hdr.count--)
      {
        if (read(upacct, &uacctbuf, (sizeof(uacct) - sizeof(uacct*))) == -1)
        {
          fprintf(stderr,"%s: error reading %s: %s\n",cmd,argv[1],sys_errlist[errno]);
          exit(1);
        } 
        if (cflag && memcmp(&uacctbuf.ac_comm, argv[1], len)) continue;
        if ((hp=gethostbyaddr((char *)&uacctbuf.addr,sizeof(struct in_addr),AF_INET)) == NULL)
          if (cflag)
            printf("%8s\t--\t% 4.2f secs\t%s\t %s",uname,(float)uacctbuf.ac_etime/(float)CLK_TCK,inet_ntoa(uacctbuf.addr),ctime(&uacctbuf.ac_btime));
          else
            printf("%8s\t--\t% 4.2f secs\t%s\t %s",&uacctbuf.ac_comm,(float)uacctbuf.ac_etime/(float)CLK_TCK,inet_ntoa(uacctbuf.addr),ctime(&uacctbuf.ac_btime));
        else
        {
        char *ptr=hp->h_name;
          while(*ptr)
          {
            if (*ptr == '.')
            {
              *ptr='\0';
              break;
            }
            ++ptr;
          }
          if (cflag)
            printf("%8s\t--\t% 4.2f secs\t%s\t%s",uname,(float)uacctbuf.ac_etime/(float)CLK_TCK,hp->h_name,ctime(&uacctbuf.ac_btime));
          else
            printf("%8s\t--\t% 4.2f secs\t%s\t%s",&uacctbuf.ac_comm,(float)uacctbuf.ac_etime/(float)CLK_TCK,hp->h_name,ctime(&uacctbuf.ac_btime));
        }
      }
      if (!cflag) break;
    }
    if (!cflag) lseek(upacct, hdr.count * (sizeof(uacct) - sizeof(uacct*)), SEEK_CUR);
  }

  /* finis */
  exit(0);
}
